import ast, os, copy

__all__ = ["api_crawler", "differ"]


class APICrawler(object):
    """
    Api Crawler: Given a directory, it will use AST to crawl through each of the
    File and produce AST nodes, which can then be parsed into a representation
    of the structure of that file. The execution...
    """

    # Exclude these folders from search.
    exclude = ("tests", "static", "sampledata", "mplexplorer")


    # Point the crawler at this directory.
    def __init__(self, directory):
        self.directory = directory

    def is_public(self, name):
        # Determines whether a given function or class is public or private. Names
        # with underscores are not allowed unless the name is "__init__"
        return not name.startswith("_", 0, 1) or (name.startswith("__") and
                name.endswith("__"))

    def is_toplevel_function(self, ast_node):
        # Checks whether a node is a top level function. Checks whether it is an AST
        # function instance. Also uses column offset, i.e. whether the function has
        # any tabs or spaces before it.
        return isinstance(ast_node, ast.FunctionDef) and ast_node.col_offset == 0

    def is_class(self, ast_node):
        # Checks whether a node is an instance of an ast Class
        return isinstance(ast_node, ast.ClassDef)

    def get_classes(self, source):
        # For a given source, walk through the file looking for classes, and assign
        # the associated methods. Pass them as a dictionary containing each class
        # and its methods.
        parsed = ast.parse(source)
        classes = [node for node in ast.walk(parsed) if self.is_class(node) and self.is_public(node.name)]
        class_defs = {}
        for x in classes:
            class_defs[x.name] = {}
            methods = []
            for y in x.body:
                if isinstance(y, ast.FunctionDef) and self.is_public(y.name):
                    methods.append(y.name)
            class_defs[x.name]["methods"] = methods
        return class_defs

    def get_functions(self, source):
        # For a given source, look for functions and return those functions as a list.
        parsed = ast.parse(source)
        functions = [node.name for node in ast.walk(parsed) if self.is_toplevel_function(node) and self.is_public(node.name)]
        return functions

    def get_filenames(self, directory):
        # Go through the file tree and grab the filnames for each file found.
        # Assign these to an array and return that array.
        files = []
        tree = os.walk(directory, topdown=True, followlinks=False)
        for dirpath, dirnames, filenames in tree:
            for folder in self.exclude:
                if folder in dirpath:
                    break
            else:
                for name in filenames:
                    if name.endswith(".py") and not name.startswith("_"):
                        files.append(os.path.join(dirpath, name))
        return files

    def get_files_dict(self, filenames):
        # Using a list of filenames, open each file and parse functions and classes
        # out of the source file.
        files_dict = {}
        for x in filenames:
            with open(x, "r") as f:
                source = f.read()
                files_dict[x] = {"classes": {}, "functions": []}
                files_dict[x]["classes"] = self.get_classes(source)
                files_dict[x]["functions"] = self.get_functions(source)
        return files_dict

    def get_crawl_dict(self):
        # Generate the crawl information using get_filenames as an argument to
        # get_dict_files
        files = self.get_filenames(self.directory)
        files_dict = self.get_files_dict(files)
        return files_dict


class Differ(object):

    def __init__(self, former, latter):
        self.former = former
        self.latter = latter
        self._additions = False
        self._operation = self.diff_operation

    @property
    def additions(self):
        return self._additions

    @additions.setter
    def additions(self, value):
        # If measuring additions instead of deletions, automatically change operator.
        self._additions = value
        if value:
            self._operation = self.combinaton_diff_operation
        else:
            self._operation = self.diff_operation


    def diff_operation(self, a, b):
        # Use this function to find the difference between two sets of dictionary
        # keys. Uses the set difference operation. Called by the diff_modules
        # function.
        return list(a - b)

    def combinaton_diff_operation(self, a, b):
        # Use a combination of symmetric_difference and difference to return a list
        # of items additions through the crawl. Allows you to get additions items instead of
        # removed items. Would produce the same result if you switched the
        # positions of the former and latter versions. Much more natural and
        # easy control using the different set operators.
        return list((a ^ b) - a)

    def diff_files(self):
        diff = {}
        intersection = {}

        intersection.update(self.latter)
        intersection.update(self.former)
        files_diff = self._operation(set(self.former), set(self.latter))

        # Diff files
        for x in intersection.keys():
            if x in files_diff:
                diff[x] = {}
        return intersection, diff

    def diff_functions_classes(self, diff, intersection):
        for x in intersection.keys():
            former_items = self.former.get(x)
            latter_items = self.latter.get(x)
            if former_items and latter_items:
                function_diff = self._operation(
                    set(former_items["functions"]),
                    set(latter_items["functions"])
                )
                class_diff = self._operation(
                    set(list(former_items["classes"].keys())),
                    set(list(latter_items["classes"].keys()))
                )
                if function_diff or list(class_diff):
                    diff[x]= copy.deepcopy(intersection[x])
                    if list(class_diff):
                        diff_dict = {y: {} for y in class_diff}
                        diff[x]["classes"] = diff_dict
                    else:
                        diff[x]["classes"] = {}

                    if function_diff:
                        diff[x]["functions"] = list(function_diff)
                    else:
                        diff[x]["functions"] = []
        return diff

    def diff_methods(self, diff, intersection):
        for x in intersection.keys():
            former_classes = self.former[x].get("classes") if self.former.get(x) else {}
            latter_classes = self.latter[x].get("classes") if self.latter.get(x) else {}
            if former_classes and latter_classes:
                for y in intersection[x]["classes"]:
                    # Prevent NoneType errors by returning empty dict.
                    former_methods = self.former[x]["classes"].get(y, {})
                    latter_methods = self.latter[x]["classes"].get(y, {})
                    if former_methods.get("methods") and latter_methods.get("methods"):
                        former_values = set(list(former_methods.values())[0])
                        latter_values = set(list(latter_methods.values())[0])
                        methods_diff = self._operation(former_values, latter_values)
                        if methods_diff:
                            if not diff.get(x):
                                diff[x] = {}
                                diff[x]["classes"] = {}
                            diff[x]["classes"][y] = copy.deepcopy(intersection[x]["classes"][y])
                            diff[x]["classes"][y]["methods"] = methods_diff
        return diff

    def diff_modules(self):
        intersection, diff = self.diff_files()
        diff = self.diff_functions_classes(diff, intersection)
        diff = self.diff_methods(diff, intersection)
        return diff

    def pretty_diff(self, diff):
        parsed_diff = []
        if self.additions:
            method = "ADDED"
        else:
            method = "DELETED"
        for x in diff.keys():
            formatted_string = "%s %s" % (method, os.path.splitext(x.replace("/", "."))[0])
            if diff[x].values():
                for y in diff[x].values():
                    if isinstance(y, dict) and y:
                        for z in y.keys():
                            if not y[z].values():
                                parsed_diff.append("%s.%s" % (formatted_string, z))
                            else:
                                for a in y[z].values():
                                    for b in a:
                                        parsed_diff.append("%s.%s.%s" % (formatted_string, z, b))
                    elif isinstance(y, list) and y:
                        for z in y:
                            class_string = "%s" % z
                            parsed_diff.append("%s.%s" % (formatted_string, class_string))
            else:
                parsed_diff.insert(0, formatted_string)
        return parsed_diff

    def get_diff(self):
        self.additions = False
        removed = self.pretty_diff(self.diff_modules())
        self.additions = True
        added = self.pretty_diff(self.diff_modules())
        return removed + added


api_crawler = APICrawler
differ = Differ



''' Represent array expressions to be computed on the client (browser) side
by BokehJS.

Expression models are useful as ``DataSpec`` values when it is desired that
the array values be computed in the browser:

.. code-block:: python

    p.circle(x={'expr': some_expression}, ...)

or using the ``expr`` convenience function:

.. code-block:: python

    from bokeh.core.properties import expr

    p.circle(x=expr(some_expression), ...)

In this case, the values of the ``x`` coordinates will be computed in the
browser by the JavaScript implementation of ``some_expression`` using a
``ColumnDataSource`` as input.

'''
from __future__ import absolute_import

from ..core.has_props import abstract
from ..core.properties import Seq, String
from ..model import Model

@abstract
class Expression(Model):
    ''' Base class for ``Expression`` models that represent a computation
    to be carried out on the client-side.

    JavaScript implementations should implement the following methods:

    .. code-block: coffeescript

        v_compute: (source) ->
            # compute an array of values

    Note that the result of this call will be automatically saved and re-used
    for each ``source`` that is passed in. If a ``source`` is changed, then
    the saved value for that source is discarded, and the next call will
    re-compute (and save) a new value. If you wish to prevent this caching, you
    may implement ``_v_compute: (source)`` instead.

    '''
    pass

class Stack(Expression):
    ''' An expression for generating arrays by summing different columns from
    a ``ColumnDataSource``.

    This expression is useful for implementing stacked bar charts at a low
    level.

    '''

    fields = Seq(String, default=[], help="""
    A sequence of fields from a ``ColumnDataSource`` to sum (elementwise). For
    example:

    .. code-block:: python

        Stack(fields=['sales', 'marketing'])

    Will compute an array of values (in the browser) by adding the elements
    of the ``'sales'`` and ``'marketing'`` columns of a data source.
    """)



from __future__ import absolute_import, print_function

from bokeh.application.handlers import FunctionHandler
from bokeh.document import Document

from bokeh.model import Model
from bokeh.core.properties import Int, Instance

class AnotherModelInTestFunction(Model):
    bar = Int(1)

class SomeModelInTestFunction(Model):
    foo = Int(2)
    child = Instance(Model)

def test_empty_func():
    def noop(doc):
        pass
    handler = FunctionHandler(noop)
    doc = Document()
    handler.modify_document(doc)
    if handler.failed:
        raise RuntimeError(handler.error)
    assert not doc.roots

def test_func_adds_roots():
    def add_roots(doc):
        doc.add_root(AnotherModelInTestFunction())
        doc.add_root(SomeModelInTestFunction())
    handler = FunctionHandler(add_roots)
    doc = Document()
    handler.modify_document(doc)
    if handler.failed:
        raise RuntimeError(handler.error)
    assert len(doc.roots) == 2

def test_safe_to_fork():
    def noop(doc):
        pass
    handler = FunctionHandler(noop)
    doc = Document()
    assert handler.safe_to_fork
    handler.modify_document(doc)
    if handler.failed:
        raise RuntimeError(handler.error)
    assert not handler.safe_to_fork



''' Provide definitions for Bokeh WebSocket prototol versions.

A *protocol specification* is a sequence of tuples of the form:

.. code-block:: python

    (
        (<message_type>, <revision>),
        (<message_type>, <revision>),
        ...
    )

Where ``<message_type>`` is string that identifies a message type, e.g,
``'ACK'``, ``'SERVER-INFO-REQ'``, etc. and ``<revision>`` is an integer that
identifies what revision of the message this version of the protocol uses.

A *protocol version* is a string of the form ``'<major>.<minor>'``. The
guidelines for updating the major or minor version are:

``<major>``
    bump when new messages are added or deleted (and reset minor
    version to zero)

``<minor>``
    bump when existing message revisions change

.. data:: spec
    :annotation:

    A mapping of protocol versions to protocol specifications.

    .. code-block:: python

        {
            "1.0" : (
                ("ACK", 1),
                ("OK", 1),
                ("ERROR", 1),
                ("EVENT", 1),
                ('SERVER-INFO-REPLY', 1),
                ('SERVER-INFO-REQ', 1),
                ('PULL-DOC-REQ', 1),
                ('PULL-DOC-REPLY', 1),
                ('PUSH-DOC', 1),
                ('PATCH-DOC', 1)
            ),
        }


'''
from __future__ import absolute_import

###############################################################################
#                                                                             #
#                                                                             #
#                                                                             #
#                       *******************************                       #
#                       ****** !!! IMPORTANT !!! ******                       #
#                       *******************************                       #
#                                                                             #
#                                                                             #
#                                                                             #
#     ANY update to this file MUST be accompanied by the "PROTOCOL" tag.      #
#                                                                             #
#                                                                             #
#                                                                             #
#                                                                             #
###############################################################################

# Please update the docstring above if anything here is changed

spec = {

    "1.0" : (
        ("ACK", 1),
        ("OK", 1),
        ("ERROR", 1),
        ("EVENT", 1),
        ('SERVER-INFO-REPLY', 1),
        ('SERVER-INFO-REQ', 1),
        ('PULL-DOC-REQ', 1),
        ('PULL-DOC-REPLY', 1),
        ('PUSH-DOC', 1),
        ('PATCH-DOC', 1)
    ),

}



''' Define the Bokeh application class.

A *Bokeh Application* is a lightweight factory for creating Bokeh Documents.
For example, whenever a new client connects to a Bokeh Server, the server uses
the ``Application`` to generate a unique new document, to service  the client
session. The application performs this task by invoking the ``modify_document``
method of any ``Handler`` objects that it is configured with.

'''

from __future__ import absolute_import

from .application import Application



from __future__ import absolute_import, print_function

import unittest

import os
import tempfile

from bokeh.document import Document
from bokeh.model import Model
from bokeh.core.property_mixins import FillProps, LineProps, TextProps
from bokeh.core.properties import Int, String
from bokeh.themes import Theme

class ThemedModel(Model):
    number = Int(42)
    string = String("hello")

class SubOfThemedModel(ThemedModel):
    another_string = String("world")

FILE_CONTENTS = """
attrs:
    ThemedModel:
        number: 57
    SubOfThemedModel:
        another_string: "boo"
""".encode('utf-8')

class TestThemes(unittest.TestCase):

    def test_construct_empty_theme_from_file(self):
        # windows will throw permissions error with auto-delete
        with (tempfile.NamedTemporaryFile(delete=False)) as file:
            # create and apply empty theme with no exception thrown
            file.file.write("".encode('utf-8'))
            file.file.flush()
            theme = Theme(filename=file.name)
            theme.apply_to_model(ThemedModel())
        file.close()
        os.remove(file.name)

    def test_construct_empty_theme_from_json(self):
        # create and apply empty theme with no exception thrown
        theme = Theme(json=dict())
        theme.apply_to_model(ThemedModel())

    def test_construct_no_json_or_filename(self):
        with self.assertRaises(ValueError) as manager:
            Theme()
        self.assertTrue("requires json or a filename" in repr(manager.exception))

    def test_construct_json_and_filename(self):
        with self.assertRaises(ValueError) as manager:
            # we check "" and {} as falsey values, to try to trick
            # our code into thinking they weren't provided.
            Theme(filename="", json={})
        self.assertTrue("not both" in repr(manager.exception))

    def test_construct_bad_attrs(self):
        with self.assertRaises(ValueError) as manager:
            Theme(json=dict(attrs=42))
        self.assertTrue("should be a dictionary of class names" in repr(manager.exception))

    def test_construct_bad_class_props(self):
        with self.assertRaises(ValueError) as manager:
            Theme(json=dict(attrs=dict(SomeClass=42)))
        self.assertTrue("should be a dictionary of properties" in repr(manager.exception))

    def test_construct_nonempty_theme_from_file(self):
        # windows will throw permissions error with auto-delete
        with (tempfile.NamedTemporaryFile(delete=False)) as file:
            # create and apply empty theme with no exception thrown
            file.file.write(FILE_CONTENTS)
            file.file.flush()
            theme = Theme(filename=file.name)
            self.assertDictEqual(dict(number=57), theme._for_class(ThemedModel))
            self.assertDictEqual(dict(number=57, another_string="boo"), theme._for_class(SubOfThemedModel))
        file.close()
        os.remove(file.name)

    def test_theming_a_model(self):
        theme = Theme(json={
            'attrs' : {
                'ThemedModel' : {
                    'string' : 'w00t'
                }
            }
        })
        obj = ThemedModel()
        changes = dict(calls=[])
        self.assertEqual('hello', obj.string)
        def record_trigger(attr, old, new_):
            changes['calls'].append((attr, old, new_))
        obj.on_change('string', record_trigger)
        theme.apply_to_model(obj)
        self.assertEqual('w00t', obj.string)
        self.assertEqual([('string', 'hello', 'w00t')], changes['calls'])

    def test_theming_a_model_via_base(self):
        theme = Theme(json={
            'attrs' : {
                'ThemedModel' : {
                    'string' : 'w00t'
                }
            }
        })
        obj = SubOfThemedModel()
        changes = dict(calls=[])
        def record_trigger(attr, old, new_):
            changes['calls'].append((attr, old, new_))
        obj.on_change('string', record_trigger)
        self.assertEqual('hello', obj.string)
        theme.apply_to_model(obj)
        self.assertEqual('w00t', obj.string)
        self.assertEqual([('string', 'hello', 'w00t')], changes['calls'])

    def test_subclass_theme_used_rather_than_base(self):
        theme = Theme(json={
            'attrs' : {
                'ThemedModel' : {
                    'string' : 'w00t'
                },
                'SubOfThemedModel' : {
                    'string' : 'bar'
                }
            }
        })
        obj = SubOfThemedModel()
        self.assertEqual('hello', obj.string)
        changes = dict(calls=[])
        def record_trigger(attr, old, new_):
            changes['calls'].append((attr, old, new_))
        obj.on_change('string', record_trigger)
        theme.apply_to_model(obj)
        self.assertEqual('bar', obj.string)
        self.assertEqual([('string', 'hello', 'bar')], changes['calls'])

    def test_theming_a_document_after_adding_root(self):
        theme = Theme(json={
            'attrs' : {
                'ThemedModel' : {
                    'string' : 'w00t'
                }
            }
        })
        obj = ThemedModel()
        doc = Document()
        doc.add_root(obj)
        self.assertEqual('hello', obj.string)
        changes = dict(calls=[])
        def record_trigger(attr, old, new_):
            changes['calls'].append((attr, old, new_))
        obj.on_change('string', record_trigger)
        doc.theme = theme
        self.assertIs(doc.theme, theme)
        self.assertEqual('w00t', obj.string)
        doc.remove_root(obj)
        self.assertEqual('hello', obj.string)
        self.assertEqual([('string', 'hello', 'w00t'),
                          ('string', 'w00t', 'hello')], changes['calls'])

    def test_theming_a_document_before_adding_root(self):
        theme = Theme(json={
            'attrs' : {
                'ThemedModel' : {
                    'string' : 'w00t'
                }
            }
        })
        obj = ThemedModel()
        doc = Document()
        self.assertEqual('hello', obj.string)
        doc.theme = theme
        self.assertIs(doc.theme, theme)
        changes = dict(calls=[])
        def record_trigger(attr, old, new_):
            changes['calls'].append((attr, old, new_))
        obj.on_change('string', record_trigger)
        doc.add_root(obj)
        self.assertEqual('w00t', obj.string)
        doc.remove_root(obj)
        self.assertEqual('hello', obj.string)
        self.assertEqual([('string', 'hello', 'w00t'),
                          ('string', 'w00t', 'hello')], changes['calls'])

    def test_setting_document_theme_to_none(self):
        theme = Theme(json={
            'attrs' : {
                'ThemedModel' : {
                    'string' : 'w00t'
                }
            }
        })
        obj = ThemedModel()
        doc = Document()
        doc.add_root(obj)
        changes = dict(calls=[])
        def record_trigger(attr, old, new_):
            changes['calls'].append((attr, old, new_))
        obj.on_change('string', record_trigger)
        doc.theme = theme
        self.assertEqual('w00t', obj.string)
        # setting to None reverts to default theme
        doc.theme = None
        self.assertIsNot(doc.theme, None)
        self.assertEqual('hello', obj.string)
        self.assertEqual([('string', 'hello', 'w00t'),
                          ('string', 'w00t', 'hello')], changes['calls'])

    def _compare_dict_to_model_class_defaults(self, props, model_class):
        model = model_class()
        for name, value in props.items():
            property = model.lookup(name)
            if property is None:
                raise RuntimeError("Model %r has no property %s" % (model, name))
            default = property.class_default(model_class)
            if default != value:
                print("%s.%s differs default %r theme %r" % (model_class.__name__, name, default, value))
            else:
                print("%s.%s default %r is identical in the theme" % (model_class.__name__, name, default))

    def _compare_dict_to_model_defaults(self, props, model_name):
        import bokeh.models as models
        import bokeh.models.widgets as widgets

        if hasattr(models, model_name):
            self._compare_dict_to_model_class_defaults(props, getattr(models, model_name))
        elif hasattr(widgets, model_name):
            self._compare_dict_to_model_class_defaults(props, getattr(widgets, model_name))
        else:
            raise RuntimeError("Could not find class for " + model_name)

    def test_default_theme_is_empty(self):
        # this is kind of a silly test once we fix default.yaml to be empty :-)
        # the point is to list all the things to fix.
        doc = Document()

        # before each assertion, we print out why it's going to fail.
        for class_name, props in doc.theme._json['attrs'].items():
            self._compare_dict_to_model_defaults(props, class_name)
        self.assertEqual(0, len(doc.theme._json['attrs']))

        self._compare_dict_to_model_class_defaults(doc.theme._fill_defaults, FillProps)
        self.assertEqual(0, len(doc.theme._fill_defaults))

        self._compare_dict_to_model_class_defaults(doc.theme._text_defaults, TextProps)
        self.assertEqual(0, len(doc.theme._text_defaults))

        self._compare_dict_to_model_class_defaults(doc.theme._line_defaults, LineProps)
        self.assertEqual(0, len(doc.theme._line_defaults))



'''

'''
from __future__ import absolute_import

from ..exceptions import ProtocolError

index = {}

def register(cls):
    ''' Decorator to add a Message (and its revision) to the Protocol index.

    Example:

        .. code-block:: python

            @register
            class some_msg_1(Message):

                msgtype  = 'SOME-MSG'
                revision = 1

                @classmethod
                def create(cls, **metadata):
                    header = cls.create_header()
                    content = {}
                    return cls(header, metadata, content)

    '''
    key = (cls.msgtype, cls.revision)
    if key in index:
        raise ProtocolError("Duplicate message specification encountered: %r" % key)
    index[key] = cls
    return cls

from .ack import *
from .event import *
from .ok import *
from .patch_doc import *
from .pull_doc_req import *
from .pull_doc_reply import *
from .push_doc import *
from .error import *
from .server_info_reply import *
from .server_info_req import *



''' Provide a request handler that returns a page displaying a document.

'''
from __future__ import absolute_import, print_function

import logging
log = logging.getLogger(__name__)

from tornado.web import StaticFileHandler

from bokeh.settings import settings

class StaticHandler(StaticFileHandler):
    ''' Implements a custom Tornado static file handler for BokehJS
    JavaScript and CSS resources.

    '''
    def __init__(self, tornado_app, *args, **kw):
        kw['path'] = settings.bokehjsdir()

        # Note: tornado_app is stored as self.application
        super(StaticHandler, self).__init__(tornado_app, *args, **kw)

    # We aren't using tornado's built-in static_path function
    # because it relies on TornadoApplication's autoconfigured
    # static handler instead of our custom one. We have a
    # custom one because we think we might want to serve
    # static files from multiple paths at once in the future.
    @classmethod
    def append_version(cls, path):
        # This version is cached on the StaticFileHandler class,
        # keyed by absolute filesystem path, and only invalidated
        # on an explicit StaticFileHandler.reset(). The reset is
        # automatic on every request if you set static_hash_cache=False
        # in TornadoApplication kwargs. In dev mode rely on dev tools
        # to manage caching. This improves the ability to debug code.
        if settings.dev:
            return path
        else:
            version = StaticFileHandler.get_version(dict(static_path=settings.bokehjsdir()), path)
            return ("%s?v=%s" % (path, version))



""" Thoroughly document Bokeh enumerations

The ``bokeh-enum`` directive generates useful type information
for the enumeration, including all the allowable values. If the
number of values is large, the full list is put in a collapsible
code block.

This directive takes the name of a Bokeh enum variable and the
module to find the value as an argument::

    .. bokeh-enum:: baz
        :module: bokeh.sphinxext.sample

Examples
--------

For the following definition of ``bokeh.sphinxext.sample.Bar``::

    baz = enumeration("a", "b", "c")

the above usage yields the output:

    .. bokeh-enum:: baz
        :module: bokeh.sphinxext.sample

"""
from __future__ import absolute_import, print_function

import importlib
import textwrap

from docutils.parsers.rst.directives import unchanged

from sphinx.errors import SphinxError

from .bokeh_directive import BokehDirective
from .templates import ENUM_DETAIL

wrapper = textwrap.TextWrapper(subsequent_indent='    ')

class BokehEnumDirective(BokehDirective):

    has_content = True
    required_arguments = 1
    optional_arguments = 2

    option_spec = {
        'module': unchanged,
        'noindex': lambda x: True, # directives.flag weirdly returns None
    }

    def run(self):

        name = self.arguments[0]

        try:
            module = importlib.import_module(self.options['module'])
        except ImportError:
            raise SphinxError("Could not generate reference docs for %r: could not import module %r" % (self.arguments[0], self.options['module']))

        enum = getattr(module, name, None)

        fullrepr = repr(enum)
        if len(fullrepr) > 180:
            shortrepr = fullrepr[:40] + " .... " + fullrepr[-40:]
            fullrepr = wrapper.wrap(fullrepr)
        else:
            shortrepr = fullrepr
            fullrepr = None

        rst_text = ENUM_DETAIL.render(
            name=name,
            module=self.options['module'],
            noindex=self.options.get('noindex', False),
            content=self.content,
            shortrepr=shortrepr,
            fullrepr=fullrepr,
        )

        return self._parse(rst_text, "<bokeh-enum>")

def setup(app):
    app.add_directive_to_domain('py', 'bokeh-enum', BokehEnumDirective)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2017, Anaconda, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest ; pytest

from bokeh.util.api import INTERNAL, PUBLIC ; INTERNAL, PUBLIC
from bokeh.util.testing import verify_api ; verify_api

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports

# External imports

# Bokeh imports
from bokeh.model import Model

# Module under test
import bokeh.embed.util as beu

#-----------------------------------------------------------------------------
# API Definition
#-----------------------------------------------------------------------------

api = {

    PUBLIC: (

    ), INTERNAL: (

        ( 'FromCurdoc',                            (1,0,0) ),
        ( 'check_models_or_docs',                  (1,0,0) ),
        ( 'check_one_model_or_doc',                (1,0,0) ),
        ( 'div_for_render_item',                   (1,0,0) ),
        ( 'find_existing_docs',                    (1,0,0) ),
        ( 'html_page_for_render_items',            (1,0,0) ),
        ( 'script_for_render_items',               (1,0,0) ),
        ( 'standalone_docs_json_and_render_items', (1,0,0) ),
        ( 'wrap_in_onload',                        (1,0,0) ),
        ( 'wrap_in_safely',                        (1,0,0) ),
        ( 'wrap_in_script_tag',                    (1,0,0) ),
        ( 'escape',                                (1,0,0) ),

    )

}

Test_api = verify_api(beu, api)

#-----------------------------------------------------------------------------
# Setup
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Public API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Internal API
#-----------------------------------------------------------------------------

class Test_FromCurdoc(object):

    def test_type(self):
        assert isinstance(beu.FromCurdoc, type)

class Test_check_models_or_docs(object):
    pass

class Test_check_one_model_or_doc(object):

    def test_succeed_with_one_model(self):
        m = Model()
        assert beu.check_one_model_or_doc(m) is m

    def test_fails_with_multiple_models(self):
        m1 = Model()
        m2 = Model()
        with pytest.raises(ValueError):
            beu.check_one_model_or_doc([m1, m2])
        with pytest.raises(ValueError):
            beu.check_one_model_or_doc((m1, m2))
        with pytest.raises(ValueError):
            beu.check_one_model_or_doc(dict(m1=m1, m2=m2))

class Test_div_for_render_item(object):

    def test_render(self):
        assert beu.div_for_render_item(dict(elementid="foo123")) == """
<div class="bk-root">
    <div class="bk-plotdiv" id="foo123"></div>
</div>"""

class Test_find_existing_docs(object):
    pass

class Test_html_page_for_render_items(object):
    pass

class Test_script_for_render_items(object):
    pass

class Test_standalone_docs_json_and_render_items(object):
    pass

class Test_wrap_in_onload(object):

    def test_render(self):
        assert beu.wrap_in_onload("code\nmorecode") == """\
(function() {
  var fn = function() {
    code
    morecode
  };
  if (document.readyState != "loading") fn();
  else document.addEventListener("DOMContentLoaded", fn);
})();\
"""

class Test_wrap_in_safely(object):

    def test_render(self):
        assert beu.wrap_in_safely("code\nmorecode") == """\
Bokeh.safely(function() {
  code
  morecode
});\
"""

class Test_wrap_in_script_tag(object):

    def test_render(self):
        assert beu.wrap_in_script_tag("code\nmorecode") == """
<script type="text/javascript">
  code
  morecode
</script>\
"""

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

def test__ONLOAD():
    assert beu._ONLOAD == """\
(function() {
  var fn = function() {
%(code)s
  };
  if (document.readyState != "loading") fn();
  else document.addEventListener("DOMContentLoaded", fn);
})();\
"""

def test__SAFELY():
    assert beu._SAFELY == """\
Bokeh.safely(function() {
%(code)s
});"""\



from __future__ import absolute_import, print_function

import unittest

from bokeh.application.handlers import CodeHandler
from bokeh.document import Document

from bokeh.util.testing import skipIfPy3

script_adds_two_roots = """
from bokeh.io import curdoc
from bokeh.model import Model
from bokeh.core.properties import Int, Instance

class AnotherModelInTestScript(Model):
    bar = Int(1)

class SomeModelInTestScript(Model):
    foo = Int(2)
    child = Instance(Model)

curdoc().add_root(AnotherModelInTestScript())
curdoc().add_root(SomeModelInTestScript())
"""

class TestCodeHandler(unittest.TestCase):

    def test_empty_script(self):
        doc = Document()
        handler = CodeHandler(source="# This script does nothing", filename="path/to/test_filename")
        handler.modify_document(doc)
        if handler.failed:
            raise RuntimeError(handler.error)

        assert not doc.roots

    @skipIfPy3("this test doesn't have a Python 3 equivalent")
    def test_exec_and___future___flags(self):
        doc = Document()
        handler = CodeHandler(source="exec(\"print \\\"XXX\\\"\")", filename="path/to/test_filename")
        handler.modify_document(doc)
        if handler.failed:
            raise RuntimeError(handler.error)

        assert not doc.roots

    def test_script_adds_roots(self):
        doc = Document()
        handler = CodeHandler(source=script_adds_two_roots, filename="path/to/test_filename")
        handler.modify_document(doc)
        if handler.failed:
            raise RuntimeError(handler.error)

        assert len(doc.roots) == 2

    def test_script_bad_syntax(self):
        doc = Document()
        handler = CodeHandler(source="This is a syntax error", filename="path/to/test_filename")
        handler.modify_document(doc)

        assert handler.error is not None
        assert 'Invalid syntax' in handler.error

    def test_script_runtime_error(self):
        doc = Document()
        handler = CodeHandler(source="raise RuntimeError('nope')", filename="path/to/test_filename")
        handler.modify_document(doc)

        assert handler.error is not None
        assert 'nope' in handler.error

    def test_script_sys_path(self):
        doc = Document()
        handler = CodeHandler(source="""import sys; raise RuntimeError("path: '%s'" % sys.path[0])""", filename="path/to/test_filename")
        handler.modify_document(doc)

        assert handler.error is not None
        assert "path: 'path/to'" in handler.error

    def test_script_argv(self):
        doc = Document()
        handler = CodeHandler(source="""import sys; raise RuntimeError("argv: %r" % sys.argv)""", filename="path/to/test_filename")
        handler.modify_document(doc)

        assert handler.error is not None
        assert "argv: ['test_filename']" in handler.error

        doc = Document()
        handler = CodeHandler(source="""import sys; raise RuntimeError("argv: %r" % sys.argv)""", filename="path/to/test_filename", argv=[10, 20, 30])
        handler.modify_document(doc)

        assert handler.error is not None
        assert "argv: ['test_filename', 10, 20, 30]" in handler.error

    def test_safe_to_fork(self):
        doc = Document()
        handler = CodeHandler(source="# This script does nothing", filename="path/to/test_filename")
        assert handler.safe_to_fork
        handler.modify_document(doc)
        if handler.failed:
            raise RuntimeError(handler.error)
        assert not handler.safe_to_fork



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2017, Anaconda, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest ; pytest

from bokeh.util.api import INTERNAL, PUBLIC ; INTERNAL, PUBLIC
from bokeh.util.testing import verify_api ; verify_api

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from mock import patch

# External imports

# Bokeh imports
from bokeh.io.state import curstate
from bokeh.resources import Resources

# Module under test
import bokeh.io.output as bio

#-----------------------------------------------------------------------------
# API Definition
#-----------------------------------------------------------------------------

api = {

    PUBLIC: (

        ( 'output_file',     (1, 0, 0) ),
        ( 'output_notebook', (1, 0, 0) ),
        ( 'reset_output',    (1, 0, 0) ),

    ), INTERNAL: (

    )

}

Test_api = verify_api(bio, api)

#-----------------------------------------------------------------------------
# Setup
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Public API
#-----------------------------------------------------------------------------

class Test_output_file(object):

    @patch('bokeh.io.state.State.output_file')
    def test_no_args(self, mock_output_file):
        default_kwargs = dict(title="Bokeh Plot", mode="cdn", root_dir=None)
        bio.output_file("foo.html")
        assert mock_output_file.call_count == 1
        assert mock_output_file.call_args[0] == ("foo.html",)
        assert mock_output_file.call_args[1] == default_kwargs

    @patch('bokeh.io.state.State.output_file')
    def test_with_args(self, mock_output_file):
        kwargs = dict(title="title", mode="cdn", root_dir="foo")
        bio.output_file("foo.html", **kwargs)
        assert mock_output_file.call_count == 1
        assert mock_output_file.call_args[0] == ("foo.html",)
        assert mock_output_file.call_args[1] == kwargs

class Test_output_notebook(object):

    @patch('bokeh.io.output.run_notebook_hook')
    def test_no_args(self, mock_run_notebook_hook):
        default_load_jupyter_args = (None, False, False, 5000)
        bio.output_notebook()
        assert mock_run_notebook_hook.call_count == 1
        assert mock_run_notebook_hook.call_args[0] == ("jupyter", "load") + default_load_jupyter_args
        assert mock_run_notebook_hook.call_args[1] == {}

    @patch('bokeh.io.output.run_notebook_hook')
    def test_with_args(self, mock_run_notebook_hook):
        load_jupyter_args = (Resources(), True, True, 1000)
        bio.output_notebook(*load_jupyter_args)
        assert mock_run_notebook_hook.call_count == 1
        assert mock_run_notebook_hook.call_args[0] == ("jupyter", "load") + load_jupyter_args
        assert mock_run_notebook_hook.call_args[1] == {}

@patch('bokeh.io.state.State.reset')
def test_reset_output(mock_reset):
    # might create a new one, which also calls reset
    original_call_count = curstate().reset.call_count
    bio.reset_output()
    assert curstate().reset.call_count == original_call_count+1

#-----------------------------------------------------------------------------
# Internal API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------



from __future__ import absolute_import

import bokeh.command.subcommands.secret as scsecret
from bokeh.command.bootstrap import main

def test_create():
    import argparse
    from bokeh.command.subcommand import Subcommand

    obj = scsecret.Secret(parser=argparse.ArgumentParser())
    assert isinstance(obj, Subcommand)

def test_name():
    assert scsecret.Secret.name == "secret"

def test_help():
    assert scsecret.Secret.help == "Create a Bokeh secret key for use with Bokeh server"

def test_args():
    assert scsecret.Secret.args == (
    )

def test_run(capsys):
    main(["bokeh", "secret"])
    out, err = capsys.readouterr()
    assert err == ""
    assert len(out) == 45
    assert out[-1] == '\n'



from __future__ import absolute_import, print_function

import pytest

from bokeh.application.handlers import ServerLifecycleHandler
from bokeh.document import Document
from bokeh.util.testing import with_file_contents

script_adds_four_handlers = """
def on_server_loaded(server_context):
    return "on_server_loaded"
def on_server_unloaded(server_context):
    return "on_server_unloaded"
def on_session_created(session_context):
    return "on_session_created"
def on_session_destroyed(session_context):
    return "on_session_destroyed"
"""

def test_empty_lifecycle():
    doc = Document()
    def load(filename):
        handler = ServerLifecycleHandler(filename=filename)
        handler.modify_document(doc)
        handler.on_server_loaded(None)
        handler.on_server_unloaded(None)
        handler.on_session_created(None)
        handler.on_session_destroyed(None)
        if handler.failed:
            raise RuntimeError(handler.error)
    with_file_contents("# This script does nothing", load)

    assert not doc.roots

def test_lifecycle_bad_syntax():
    result = {}
    def load(filename):
        handler = ServerLifecycleHandler(filename=filename)
        result['handler'] = handler
    with_file_contents("This is a syntax error", load)

    handler = result['handler']
    assert handler.error is not None
    assert 'Invalid syntax' in handler.error

def test_lifecycle_runtime_error():
    result = {}
    def load(filename):
        handler = ServerLifecycleHandler(filename=filename)
        result['handler'] = handler
    with_file_contents("raise RuntimeError('nope')", load)

    handler = result['handler']
    assert handler.error is not None
    assert 'nope' in handler.error

def test_lifecycle_bad_server_loaded_signature():
    result = {}
    def load(filename):
        handler = ServerLifecycleHandler(filename=filename)
        result['handler'] = handler
    with_file_contents("""
def on_server_loaded(a,b):
    pass
""", load)

    handler = result['handler']
    assert handler.error is not None
    assert 'on_server_loaded must have signature func(server_context)' in handler.error
    assert 'func(a, b)' in handler.error
    assert "Traceback" in handler.error_detail

def test_lifecycle_bad_server_unloaded_signature():
    result = {}
    def load(filename):
        handler = ServerLifecycleHandler(filename=filename)
        result['handler'] = handler
    with_file_contents("""
def on_server_unloaded(a,b):
    pass
""", load)

    handler = result['handler']
    assert handler.error is not None
    assert 'on_server_unloaded must have signature func(server_context)' in handler.error
    assert 'func(a, b)' in handler.error
    assert "Traceback" in handler.error_detail

def test_lifecycle_bad_session_created_signature():
    result = {}
    def load(filename):
        handler = ServerLifecycleHandler(filename=filename)
        result['handler'] = handler
    with_file_contents("""
def on_session_created(a,b):
    pass
""", load)

    handler = result['handler']
    assert handler.error is not None
    assert 'on_session_created must have signature func(session_context)' in handler.error
    assert 'func(a, b)' in handler.error

def test_lifecycle_bad_session_destroyed_signature():
    result = {}
    def load(filename):
        handler = ServerLifecycleHandler(filename=filename)
        result['handler'] = handler
    with_file_contents("""
def on_session_destroyed(a,b):
    pass
""", load)

    handler = result['handler']
    assert handler.error is not None
    assert 'on_session_destroyed must have signature func(session_context)' in handler.error
    assert 'func(a, b)' in handler.error

def test_calling_lifecycle_hooks():
    result = {}
    def load(filename):
        handler = result['handler'] = ServerLifecycleHandler(filename=filename)
        if handler.failed:
            raise RuntimeError(handler.error)
    with_file_contents(script_adds_four_handlers, load)

    handler = result['handler']
    assert "on_server_loaded" == handler.on_server_loaded(None)
    assert "on_server_unloaded" == handler.on_server_unloaded(None)
    assert "on_session_created" == handler.on_session_created(None)
    assert "on_session_destroyed" == handler.on_session_destroyed(None)

def test_missing_filename_raises():
    with pytest.raises(ValueError):
        ServerLifecycleHandler()

def test_url_path():
    result = {}
    def load(filename):
        handler = ServerLifecycleHandler(filename=filename)
        result['handler'] = handler
    with_file_contents("""
def on_server_unloaded(server_context):
    pass
""", load)

    handler = result['handler']
    assert handler.error is None
    assert handler.url_path().startswith("/")

def test_url_path_failed():
    result = {}
    def load(filename):
        handler = ServerLifecycleHandler(filename=filename)
        result['handler'] = handler
    with_file_contents("""
# bad signature
def on_server_unloaded():
    pass
""", load)

    handler = result['handler']
    assert handler.error is not None
    assert handler.url_path() is None



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2017, Anaconda, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
'''

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
log = logging.getLogger(__name__)

from bokeh.util.api import public, internal ; public, internal

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from contextlib import contextmanager

# External imports

# Bokeh imports
from ..core.templates import DOC_JS
from ..core.json_encoder import serialize_json
from ..settings import settings
from ..util.string import encode_utf8
from .util import FromCurdoc
from .util import check_one_model_or_doc, div_for_render_item, find_existing_docs, standalone_docs_json_and_render_items

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Public API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Internal API
#-----------------------------------------------------------------------------

@internal((1,0,0))
def notebook_content(model, notebook_comms_target=None, theme=FromCurdoc):
    ''' Return script and div that will display a Bokeh plot in a Jupyter
    Notebook.

    The data for the plot is stored directly in the returned HTML.

    Args:
        model (Model) : Bokeh object to render

        notebook_comms_target (str, optional) :
            A target name for a Jupyter Comms object that can update
            the document that is rendered to this notebook div

        theme (Theme, optional) :
            Defaults to the ``Theme`` instance in the current document.
            Setting this to ``None`` uses the default theme or the theme
            already specified in the document. Any other value must be an
            instance of the ``Theme`` class.

    Returns:
        script, div, Document

    .. note::
        Assumes :func:`~bokeh.io.notebook.load_notebook` or the equivalent
        has already been executed.

    '''

    model = check_one_model_or_doc(model)

    # Comms handling relies on the fact that the new_doc returned here
    # has models with the same IDs as they were started with
    with _ModelInEmptyDocument(model, apply_theme=theme) as new_doc:
        (docs_json, render_items) = standalone_docs_json_and_render_items([model])

    item = render_items[0]
    if notebook_comms_target:
        item['notebook_comms_target'] = notebook_comms_target
    else:
        notebook_comms_target = ''

    script = DOC_JS.render(
        docs_json=serialize_json(docs_json),
        render_items=serialize_json(render_items)
    )

    div = div_for_render_item(item)

    return encode_utf8(script), encode_utf8(div), new_doc

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

@contextmanager
def _ModelInEmptyDocument(model, apply_theme=None):

    # Note: Comms handling relies on the fact that the new_doc returned
    # has models with the same IDs as they were started with

    from ..document import Document
    doc = find_existing_docs([model])

    if apply_theme is FromCurdoc:
        from ..io import curdoc; curdoc
        doc.theme = curdoc().theme
    elif apply_theme is not None:
        doc.theme = apply_theme

    model._document = None
    for ref in model.references():
        ref._document = None
    new_doc = Document()
    new_doc.add_root(model)

    if settings.perform_document_validation():
        new_doc.validate()

    yield new_doc

    model._document = doc
    for ref in model.references():
        ref._document = doc

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



from __future__ import absolute_import, print_function

import argparse
import contextlib
import re
import socket
import subprocess
import sys

import pytest
import requests

import bokeh.command.subcommands.serve as scserve

from bokeh.resources import DEFAULT_SERVER_PORT



def test_create():
    import argparse
    from bokeh.command.subcommand import Subcommand

    obj = scserve.Serve(parser=argparse.ArgumentParser())
    assert isinstance(obj, Subcommand)

def test_loglevels():
    assert scserve.LOGLEVELS == ('trace', 'debug', 'info', 'warning', 'error', 'critical')

def test_name():
    assert scserve.Serve.name == "serve"

def test_help():
    assert scserve.Serve.help == "Run a Bokeh server hosting one or more applications"

def test_args():
    from bokeh.util.string import nice_join

    assert scserve.Serve.args == (
        ('--port', dict(
            metavar = 'PORT',
            type    = int,
            help    = "Port to listen on",
            default = DEFAULT_SERVER_PORT
        )),

        ('--address', dict(
            metavar = 'ADDRESS',
            type    = str,
            help    = "Address to listen on",
            default = None,
        )),

        ('--log-level', dict(
            metavar = 'LOG-LEVEL',
            action  = 'store',
            default = 'info',
            choices = scserve.LOGLEVELS,
            help    = "One of: %s" % nice_join(scserve.LOGLEVELS),
        )),

        ('--log-format', dict(
            metavar ='LOG-FORMAT',
            action  = 'store',
            default = scserve.DEFAULT_LOG_FORMAT,
            help    = "A standard Python logging format string (default: %r)" % scserve.DEFAULT_LOG_FORMAT.replace("%", "%%"),
        )),

        ('files', dict(
            metavar='DIRECTORY-OR-SCRIPT',
            nargs='*',
            help="The app directories or scripts to serve (serve empty document if not specified)",
            default=None,
        )),

        ('--args', dict(
            metavar='COMMAND-LINE-ARGS',
            nargs=argparse.REMAINDER,
            help="Any command line arguments remaining are passed on to the application handler",
        )),

        ('--show', dict(
            action='store_true',
            help="Open server app(s) in a browser",
        )),

        ('--allow-websocket-origin', dict(
            metavar='HOST[:PORT]',
            action='append',
            type=str,
            help="Public hostnames which may connect to the Bokeh websocket",
        )),

        ('--host', dict(
            metavar='HOST[:PORT]',
            action='append',
            type=str,
            help="*** DEPRECATED ***",
        )),

        ('--prefix', dict(
            metavar='PREFIX',
            type=str,
            help="URL prefix for Bokeh server URLs",
            default=None,
        )),

        ('--keep-alive', dict(
            metavar='MILLISECONDS',
            type=int,
            help="How often to send a keep-alive ping to clients, 0 to disable.",
            default=None,
        )),

        ('--check-unused-sessions', dict(
            metavar='MILLISECONDS',
            type=int,
            help="How often to check for unused sessions",
            default=None,
        )),

        ('--unused-session-lifetime', dict(
            metavar='MILLISECONDS',
            type=int,
            help="How long unused sessions last",
            default=None,
        )),

        ('--stats-log-frequency', dict(
            metavar='MILLISECONDS',
            type=int,
            help="How often to log stats",
            default=None,
        )),

        ('--use-xheaders', dict(
            action='store_true',
            help="Prefer X-headers for IP/protocol information",
        )),

        ('--session-ids', dict(
            metavar='MODE',
            action  = 'store',
            default = None,
            choices = scserve.SESSION_ID_MODES,
            help    = "One of: %s" % nice_join(scserve.SESSION_ID_MODES),
        )),

        ('--disable-index', dict(
            action = 'store_true',
            help    = 'Do not use the default index on the root path',
        )),

        ('--disable-index-redirect', dict(
            action = 'store_true',
            help    = 'Do not redirect to running app from root path',
        )),

        ('--num-procs', dict(
             metavar='N',
             action='store',
             help="Number of worker processes for an app. Default to one. Using "
                  "0 will autodetect number of cores",
             default=1,
             type=int,
         )),
    )


@contextlib.contextmanager
def run_bokeh_serve(args):
    cmd = [sys.executable, "-m", "bokeh", "serve"] + args
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    try:
        yield p
    except Exception:
        p.terminate()
        p.wait()
        print("---- An error occurred, subprocess stdout follows ----")
        print(p.stdout.read().decode())
        raise
    else:
        p.terminate()
        p.wait()

def check_error(args):
    cmd = [sys.executable, "-m", "bokeh", "serve"] + args
    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        assert e.returncode == 1
        out = e.output.decode()
    else:
        pytest.fail("command %s unexpected successful" % (cmd,))
    return out

def test__fixup_deprecated_host_args():
    from argparse import Namespace

    args = Namespace(host=None, allow_websocket_origin=None)
    scserve._fixup_deprecated_host_args(args)
    assert args.allow_websocket_origin == None

    args = Namespace(host=[], allow_websocket_origin=None)
    scserve._fixup_deprecated_host_args(args)
    assert args.allow_websocket_origin == None

    args = Namespace(host=[], allow_websocket_origin=[])
    scserve._fixup_deprecated_host_args(args)
    assert set(args.allow_websocket_origin) == set([])

    args = Namespace(host=['*'], allow_websocket_origin=[])
    scserve._fixup_deprecated_host_args(args)
    assert set(args.allow_websocket_origin) == set(['*'])

    args = Namespace(host=['*'], allow_websocket_origin=['*'])
    scserve._fixup_deprecated_host_args(args)
    assert set(args.allow_websocket_origin) == set(['*'])

    args = Namespace(host=['*'], allow_websocket_origin=['*', 'foo'])
    scserve._fixup_deprecated_host_args(args)
    assert set(args.allow_websocket_origin) == set(['*', 'foo'])

    args = Namespace(host=[], allow_websocket_origin=['*'])
    scserve._fixup_deprecated_host_args(args)
    assert set(args.allow_websocket_origin) == set(['*'])

    args = Namespace(host=[], allow_websocket_origin=['*', 'foo'])
    scserve._fixup_deprecated_host_args(args)
    assert set(args.allow_websocket_origin) == set(['*', 'foo'])

    args = Namespace(host=None, allow_websocket_origin=['*'])
    scserve._fixup_deprecated_host_args(args)
    assert set(args.allow_websocket_origin) == set(['*'])

    args = Namespace(host=None, allow_websocket_origin=['*', 'foo'])
    scserve._fixup_deprecated_host_args(args)
    assert set(args.allow_websocket_origin) == set(['*', 'foo'])

def test_host_not_available():
    host = "8.8.8.8"
    out = check_error(["--address", host])
    expected = "Cannot start Bokeh server, address %r not available" % host
    assert expected in out

def test_port_not_available():
    sock = socket.socket()
    try:
        sock.bind(('0.0.0.0', 0))
        port = sock.getsockname()[1]
        out = check_error(["--port", str(port)])
        expected = "Cannot start Bokeh server, port %d is already in use" % port
        assert expected in out
    finally:
        sock.close()

def test_actual_port_printed_out():
    with run_bokeh_serve(["--port", "0"]) as p:
        pat = re.compile(r'Bokeh app running at: http://localhost:(\d+)')
        while True:
            line = p.stdout.readline()
            print("child stdout>", line)
            m = pat.search(line.decode())
            if m is not None:
                break
        else:
            pytest.fail("no matching log line in process output")
        port = int(m.group(1))
        assert port > 0
        r = requests.get("http://localhost:%d/" % (port,))
        assert r.status_code == 200



''' Provide Bokeh-specific warning subclasses.

The primary use of these subclasses to to force them to be unconditionally
displayed to users by default.

'''

class BokehDeprecationWarning(DeprecationWarning):
    ''' A Bokeh-specific ``DeprecationWarning`` subclass.

    Used to selectively filter Bokeh deprecations for unconditional display.

    '''

class BokehUserWarning(UserWarning):
    ''' A Bokeh-specific ``UserWarning`` subclass.

    Used to selectively filter Bokeh warnings for unconditional display.

    '''



from __future__ import absolute_import, print_function

from bokeh.server.server import Server
from tornado.ioloop import IOLoop
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.websocket import websocket_connect

def url(server, prefix=""):
    return "http://localhost:" + str(server.port) + prefix + "/"

def ws_url(server, prefix=""):
    return "ws://localhost:" + str(server.port) + prefix + "/ws"

def http_get(io_loop, url):
    result = {}
    def handle_request(response):
        result['response'] = response
        io_loop.stop()

    # for some reason passing a loop to AsyncHTTPClient is deprecated
    assert io_loop is IOLoop.current()
    http_client = AsyncHTTPClient()
    headers = dict()
    http_client.fetch(url, handle_request, headers=headers)
    io_loop.start()

    if 'response' not in result:
        raise RuntimeError("Failed to http get")
    response = result['response']
    if response.error:
        raise response.error
    else:
        return response

def websocket_open(io_loop, url, origin=None):
    result = {}
    def handle_connection(future):
        result['connection'] = future
        io_loop.stop()

    request = HTTPRequest(url)
    if origin is not None:
        request.headers['Origin'] = origin
    websocket_connect(request, callback=handle_connection, io_loop=io_loop)

    io_loop.start()

    if 'connection' not in result:
        raise RuntimeError("Failed to handle websocket connect")
    future = result['connection']
    if future.exception():
        raise future.exception()
    else:
        future.result().close()
        return None

# lets us use a current IOLoop with "with"
# and ensures the server unlistens
class ManagedServerLoop(object):
    def __init__(self, application, **server_kwargs):
        loop = IOLoop()
        loop.make_current()
        server_kwargs['io_loop'] = loop
        self._server = Server(application, **server_kwargs)
    def __exit__(self, type, value, traceback):
        self._server.unlisten()
        self._server.stop()
        self._server.io_loop.close()
    def __enter__(self):
        self._server.start()
        return self._server
    @property
    def io_loop(self):
        return self.s_server.io_loop



from __future__ import absolute_import

from bokeh import __version__

from ..message import Message
from . import register

_VERSION_INFO = {
    'bokeh'  : __version__,
    'server' : __version__,
}

@register
class server_info_reply_1(Message):
    ''' Define the ``SERVER-INFO-REPLY`` message (revision 1) for replying to
    Server info requests from clients.

    The ``content`` fragment of for this message is has the form:

    .. code-block:: python

        {
            'version_info' : {
                'bokeh'  : <bokeh library version>
                'server' : <bokeh server version>
            }
        }

    '''

    msgtype  = 'SERVER-INFO-REPLY'
    revision = 1

    @classmethod
    def create(cls, request_id, **metadata):
        ''' Create an ``SERVER-INFO-REPLY`` message

        Args:
            request_id (str) :
                The message ID for the message that issues the info request

        Any additional keyword arguments will be put into the message
        ``metadata`` fragment as-is.

        '''
        header = cls.create_header(request_id=request_id)
        content = {
            'version_info': _VERSION_INFO,
        }
        return cls(header, metadata, content)
